/*
  DCF77.c - DCF77 library 
  Copyright (c) Thijs Elenbaas 2012

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
  11 Apr 2012 - initial release 
  23 Apr 2012 - added UTC support
  2  Jul 2012 - minor bugfix and additional noise rejection
*/

#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include "stm32l0xx_hal.h"
#include "dcf77.h"

// Parameters shared between interupt loop and main loop
volatile uint64_t filledBuffer = 0;
volatile uint8_t FilledBufferAvailable = 0;
volatile time_t filledTimestamp = 0;

// DCF Buffers and indicators
int16_t bufferPosition = 0;
uint64_t runningBuffer = 0;
uint64_t processingBuffer = 0;

// Pulse flanks
uint32_t leadingEdge = 0;
uint32_t trailingEdge = 0;
uint32_t PreviousLeadingEdge = 0;
uint8_t Up = 0;

// DCF77 and internal timestamps
time_t latestupdatedTime = 0;
time_t previousUpdatedTime = 0;
time_t processingTimestamp = 0;
time_t previousProcessingTimestamp = 0;
uint8_t CEST = 0;
dcf77_buffer_t DCF77Buffer;
dcf77_flags_t flags = {0,0,0,0};

//Enable flag
uint8_t dcf77_enabled = 0;

//Private functions
void DCF77_bufferinit(void);
void DCF77_finalizeBuffer(void);
uint8_t DCF77_receivedTimeUpdate(void);
void DCF77_storePreviousTime(void);
void DCF77_calculateBufferParities(void);
uint8_t DCF77_processBuffer(void);
void DCF77_appendSignal(uint8_t signal);


void DCF77_Init(void) 
{	
    leadingEdge           = 0;
    trailingEdge          = 0;
    PreviousLeadingEdge   = 0;
    Up                    = 0;
    runningBuffer		  = 0;
    FilledBufferAvailable = 0;
    bufferPosition        = 0;
    flags.parityDate      = 0;
    flags.parityFlag      = 0;
    flags.parityHour      = 0;
    flags.parityMin       = 0;
    CEST				  = 0;
}

/**
 * Start receiving DCF77 information
 */
void DCF77_Start(void) 
{
	dcf77_enabled = 1;
}

/**
 * Stop receiving DCF77 information
 */
void DCF77_Stop(void) 
{
    dcf77_enabled = 0;
}

/**
 * Initialize buffer for next time update
 */
inline void DCF77_bufferinit(void) 
{
	runningBuffer    = 0;
	bufferPosition   = 0;
}

/**
 * Interrupt handler that processes up-down flanks into pulses and stores these in the buffer
 */
 void DCF77_intHandler(void) {
    uint32_t flankTime;
    int16_t difference;
    uint8_t sensorValue;
    
    if (!dcf77_enabled) return;
    
    flankTime = HAL_GetTick();
    sensorValue = DCF77_INPUT_PIN;

    // If flank is detected quickly after previous flank up
    // this will be an incorrect pulse that we shall reject
    if (((uint32_t)(flankTime-PreviousLeadingEdge))<DCFRejectionTime) {
        //printf("rCT\r\n");
        return;
    }

    // If the detected pulse is too short it will be an
    // incorrect pulse that we shall reject as well
    if (((uint32_t)(flankTime-leadingEdge))<DCFRejectPulseWidth) {
        //printf("rPW\r\n");
        return;
    }

    if(sensorValue == PULSE_START) {
        if (!Up) {
            // Flank up
            leadingEdge=flankTime;
            Up = 1;		                
        } 
    } else {
        if (Up) {
            // Flank down
            trailingEdge=flankTime;
            difference=(uint32_t)(trailingEdge - leadingEdge);            

            if (((uint32_t)(leadingEdge-PreviousLeadingEdge)) > DCFSyncTime) {
                DCF77_finalizeBuffer();
            }         
            PreviousLeadingEdge = leadingEdge;       
            // Distinguish between long and short pulses
            if (difference < DCFSplitTime) { DCF77_appendSignal(0); } else { DCF77_appendSignal(1); }
            Up = 0;	 
        } 
    }
}

/**
 * Add new bit to buffer
 */
inline void DCF77_appendSignal(uint8_t signal) {
	//printf("%d\r\n", signal);
	runningBuffer = runningBuffer | ((uint64_t) signal << bufferPosition);  
	bufferPosition++;
	if (bufferPosition > 59) {
		// Buffer is full before at end of time-sequence 
		// this may be due to noise giving additional peaks
		//printf("EoB\r\n");
		DCF77_finalizeBuffer();
	}
}

/**
 * Finalize filled buffer
 */
inline void DCF77_finalizeBuffer(void) {
  if (bufferPosition == 59) {
		// Buffer is full
		//printf("BF\r\n");
		// Prepare filled buffer and time stamp for main loop
		filledBuffer = runningBuffer;    
		filledTimestamp = time(NULL);
		// Reset running buffer
		DCF77_bufferinit();
		FilledBufferAvailable = 1;    
    } else {
		// Buffer is not yet full at end of time-sequence
		//printf("EoM\r\n");
		// Reset running buffer
		DCF77_bufferinit();      
    }
}

/**
 * Returns whether there is a new time update available
 * This functions should be called prior to getTime() function.
 */
uint8_t DCF77_receivedTimeUpdate(void) {
    time_t processedTime;
	// If buffer is not filled, there is no new time
	if(!FilledBufferAvailable) {
		return 0;
	}
	// if buffer is filled, we will process it and see if this results in valid parity
	if (!DCF77_processBuffer()) {
		//printf("DCF: Invalid parity\r\n");
		return 0;
	}
    
	// Since the received signal is error-prone, and the parity check is not very strong, 
	// we will do some sanity checks on the time
	processedTime = latestupdatedTime + (time(NULL) - processingTimestamp);
    //printf("DFC: Rrocessed time: %lu\r\n", processedTime);
	if (processedTime<MIN_TIME || processedTime>MAX_TIME) {
		//printf("DCF: Time outside of bounds\r\n");
		return 0;
	}

	// If received time is close to internal clock (2 min) we are satisfied
	time_t difference = abs(processedTime - time(NULL));
	if(difference < 2*SECS_PER_MIN) {
		//printf("DCF: close to internal clock\r\n");
		DCF77_storePreviousTime();
		return 1;
	}

	// Time can be further from internal clock for several reasons
	// We will check if lag from internal clock is consistent
	time_t shiftPrevious = (previousUpdatedTime - previousProcessingTimestamp);
	time_t shiftCurrent = (latestupdatedTime - processingTimestamp);	
	time_t shiftDifference = abs(shiftCurrent-shiftPrevious);
	DCF77_storePreviousTime();
	if(shiftDifference < 2*SECS_PER_MIN) {
		//printf("DCF: time lag consistent\r\n");
		return 1;
	} else {
		//printf("DCF: time lag inconsistent\r\n");
	}
	
	// If lag is inconsistent, this may be because of no previous stored date 
	// This would be resolved in a second run.
	return 0;
}

/**
 * Store previous time. Needed for consistency 
 */
void DCF77_storePreviousTime(void) {
	previousUpdatedTime = latestupdatedTime;
	previousProcessingTimestamp = processingTimestamp;
}

/**
 * Calculate the parity of the time and date. 
 */
void DCF77_calculateBufferParities(void) {
    uint8_t s;
    int16_t pos;
    
	// Calculate Parity 
	flags.parityFlag = 0;	
	for(pos=0;pos<59;pos++) {
		s = (processingBuffer >> pos) & 1;  
		
		// Update the parity bits. First: Reset when minute, hour or date starts.
		if (pos ==  21 || pos ==  29 || pos ==  36) {
			flags.parityFlag = 0;
		}
		// save the parity when the corresponding segment ends
		if (pos ==  28) {flags.parityMin = flags.parityFlag;};
		if (pos ==  35) {flags.parityHour = flags.parityFlag;};
		if (pos ==  58) {flags.parityDate = flags.parityFlag;};
		// When we received a 1, toggle the parity flag
		if (s == 1) {
			flags.parityFlag = flags.parityFlag ^ 1;
		}
	}
}

/**
 * Evaluates the information stored in the buffer. This is where the DCF77
 * signal is decoded 
 */
uint8_t DCF77_processBuffer(void) {	
//	uint8_t proccessedSucces;
    dcf77_buffer_t *rx_buffer;
    struct tm time;
    
	/////  Start interaction with interrupt driven loop  /////
	
	// Copy filled buffer and timestamp from interrupt driven loop
	processingBuffer = filledBuffer;
	processingTimestamp = filledTimestamp;
	// Indicate that there is no filled, unprocessed buffer anymore
	FilledBufferAvailable = 0;  
	
	/////  End interaction with interrupt driven loop   /////

	//  Calculate parities for checking buffer
	DCF77_calculateBufferParities();
	
	rx_buffer = (dcf77_buffer_t *)&processingBuffer;

	// Check parities
    if (flags.parityMin == rx_buffer->P1  &&
        flags.parityHour == rx_buffer->P2 &&
        flags.parityDate == rx_buffer->P3 &&
		rx_buffer->CEST != rx_buffer->CET) 
    { 
      //convert the received buffer into time

      time.tm_sec = 0;
	  time.tm_min = rx_buffer->Min-((rx_buffer->Min/16)*6);
      time.tm_hour   = rx_buffer->Hour-((rx_buffer->Hour/16)*6);
      time.tm_mday    = rx_buffer->Day-((rx_buffer->Day/16)*6); 
      time.tm_mon  = (rx_buffer->Month-((rx_buffer->Month/16)*6))-1;
      time.tm_year   = 2000 + rx_buffer->Year-((rx_buffer->Year/16)*6) -1900;
	  latestupdatedTime = mktime(&time);	 
	  CEST = rx_buffer->CEST;
	  //Parity correct
	  return 1;
	} else {
	  //Parity incorrect
	  return 0;
	}
}


//Get most recently received time 
//Note, this only returns an time once, until the next update
time_t DCF77_getTime(void)
{
    time_t currentTime;
	if (!DCF77_receivedTimeUpdate()) {
		return(0);
	} else {
		// Send out time, taking into account the difference between when the DCF time was received and the current time
		currentTime =latestupdatedTime + (time(NULL) - processingTimestamp);
		return(currentTime);
	}
}


//Get most recently received time in UTC 
//Note, this only returns an time once, until the next update
time_t DCF77_getUTCTime(void)
{
    time_t currentTime;
	if (!DCF77_receivedTimeUpdate()) {
		return(0);
	} else {
		// Send out time UTC time
		int UTCTimeDifference = (CEST ? 2 : 1)*SECS_PER_HOUR;
		currentTime =latestupdatedTime - UTCTimeDifference + (time(NULL) - processingTimestamp);
		return(currentTime);
	}
}

